//*********************************************************************************
// class TestSolarSystem
//*********************************************************************************
import java.util.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.awt.event.*;
import java.awt.*;
import java.text.NumberFormat;
import java.lang.reflect.Field;

import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.universe.SimpleUniverse;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.behaviors.mouse.*;
import com.sun.j3d.utils.behaviors.keyboard.*;
import com.sun.j3d.utils.behaviors.vp.*;
import com.sun.j3d.utils.picking.*;
import com.sun.j3d.utils.picking.behaviors.*;
import com.sun.j3d.utils.image.TextureLoader;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.loaders.*;
import com.sun.j3d.loaders.objectfile.*;

/**
@todo Planets and sun visibility back and front clip distances.
@todo rotators activation limits.
@todo remove moon from main planet array, and treat as normal satellite
*/
public class TestSolarSystem{

  //=================================================================================
  //                            INSTANCE VARIABLES
  //=================================================================================
	SimpleUniverse	 		_universe;
	Sphere[]				_planets;
	RotationInterpolator[] 	_revolutionators;
	TransformGroup			_viewerTG;
	// GUI components
	JFrame 					_frame;
	Canvas3D 				_canvas, _canvas1, _canvas2;
	JButton 				_btnPlay, _btnPause, _btnGo;
	JComboBox 				_cmbGo;
  JCheckBox				_chkFromAbove;
	JTextArea				_logArea;
	JLabel					_coordLabel;
	

  //=================================================================================
  //                            CONSTANTS
  //=================================================================================

	private static final boolean DEBUG			= false;

	private static final Color3f YELLOW 		= new Color3f(Color.yellow);
	private static final Color3f WHITE 			= new Color3f(Color.white);

	private static final String FS 				= System.getProperty("file.separator");
	private static final String LS 				= System.getProperty("line.separator");

	// directory where the textuires are stored (includes last file separator)
	private static final String TEXTURE_DIR = ".." + FS + "data" + FS + "planets" + FS;

	private static final int CANVAS1	= 0;
	private static final int CANVAS2	= 1;

	// Planets indices - used for array purposes - must be 0, 1, ...
	private static final int MERCURY 		= 0;
	private static final int VENUS 			= 1;
	private static final int EARTH 			= 2;
	private static final int MARS 			= 3;
	private static final int JUPITER 		= 4;
	private static final int SATURN 		= 5;
	private static final int URANUS 		= 6;
	private static final int NEPTUNE 		= 7;
	private static final int PLUTO 			= 8;
	private static final int SUN 			= 9;
	private static final int MOON 			= 10;
	private static final int NB_PLANETS 	= 11;

  /** Value of an astronomical unit, in km (value IERS 1992). */
	private static final double KM_PER_AU = 149597870.61;

	//************************************************************
	// 		Planetary data (check inclination of moon)
	//************************************************************

	private static final String[] PLANET_NAMES = {"Mercure", "Vnus", "Terre", "Mars", "Jupiter", "Saturne", "Uranus", "Neptune", "Pluton", "Soleil", "Lune"};

	// Radius of the planets, in km.
	private static final double[] PLANET_RADIUS = {2439, 6052, 6378, 3393.5, 71398, 60330, 25559, 24300, 1150, 695500, 1737.4};

	// Semi major axis, in astronomical unit (u.a) - geocentric value for the moon
	private static final double[] PLANET_SEMIAXIS = {0.39, 0.723, 1, 1.524, 5.203, 9.539, 19.18, 30.06, 39.53, 0.0, 0.00257};

	// Revolution periods, in terrestrial years - Sun's value meaningless
	private static final double[] PLANET_REVOLUTIONS = {0.241, 0.615, 1, 1.88, 11.862, 29.456, 84.07, 164.81, 247.7, 0.0, 0.0748};

	// Rotation periods, in terrestrial days - Average value taken for the sun (between 25.6 and 36 days)
	private static final double[] PLANET_ROTATIONS = {58.7, 243, 1, 1.026, 0.41, 0.425, 0.7459, 0.7959, 6.39, 30.0, 27.32};

	// Inclination of axis compared with the planet's orbit, in decimal degrees
	// source : http://solarsystem.nasa.gov/features/planets/venus/venus.html 'n' co
	private static final double[] PLANET_INCLINATION = {0.0, 177.3, 23.45, 25.19, 3.12, 26.73, 97.86, 29.58, 119.61, 7.25, 0.0};

	// Initial positions of the planets (value : constant term of mean longitude for J2000) - Sun's and moon's values meaningless
	private static final double[] PLANET_LAMBDAS = {252.25, 181.98, 100.47, 355.43, 34.35, 50.07, 314.06, 304.35, 245.00, 0.0, 0.0};

	//************************************************************
	// 					Jovian satellites data
	//************************************************************
	private static final String[] JUPSAT_NAMES = {"Io", "Europe", "Ganymde", "Callisto"};

	// Radius of the jovian satellites, in km.
	private static final double[] JUPSAT_RADIUS = {1821.3, 1565, 2634, 2403};

	// Semi major axis, in jovian radii
	private static final double[] JUPSAT_SEMIAXIS = {5.905, 9.397, 14.99, 26.37};

	// Revolution periods, in terrestrial days
	private static final double[] JUPSAT_REVOLUTIONS = {1.769138, 3.551181, 7.154553, 16.689018};

	// Rotation periods, in terrestrial days
	private static final double[] JUPSAT_ROTATIONS = {1.769138, 3.551181, 7.154553, 16.689018};

	// Bounding Sphere for big influence
	private static final BoundingSphere BS = new BoundingSphere(new Point3d (), getSemiAxis(PLUTO)*2f);
	// Bounding Sphere for small influence
	private static BoundingSphere BS2 = new BoundingSphere(new Point3d (), getSemiAxis(MERCURY)/50f);

  //=================================================================================
  //               CONSTANTS DEPENDING ON CONFIG FILE
  //=================================================================================

    // Which planet rotation axis should be represented ?
	private static boolean[] DRAW_AXIS = new boolean[NB_PLANETS];

	//************************************************************
	// 				Scale factors (initialized in loadConfig)
	//************************************************************
	
	// Factor by which real planet radius in km must be divided to create the virtual universe
	private static double SCALE_RADIUS;
	
	// Factor by which real distances in km must be divided to create the virtual universe
	private static double SCALE_SEMIAXIS;
	
	// Factor by which real rotation periods must be divided to create the virtual universe (result in milliseconds)
	private static double SCALE_TIME_ROTATION;
	
	// Factor by which real revolution periods must be divided to create the virtual universe (result in milliseconds)
	private static double SCALE_TIME_REVOLUTION;

	// Cheat factors - see config file for description
	private static double[] 	CHEAT_RADIUS = new double[NB_PLANETS]; // for planets
	private static double 		CHEAT_MOON_RADIUS;
	private static double 		CHEAT_MOON_SEMIAXIS;
	private static double 		CHEAT_JUPSAT_RADIUS;
	private static double 		CHEAT_JUPSAT_SEMIAXIS;
	private static double 		CHEAT_CHARON_RADIUS;
	private static double 		CHEAT_CHARON_SEMIAXIS;
	static{loadConfig();};

	// Initial position of the eye for canvas1
	private static final Point3d INITIAL_POS1 = new Point3d(8 * PLANET_RADIUS[SUN]/SCALE_RADIUS, 0, 0);

  //=================================================================================
  //                                      METHODS
  //=================================================================================

	//****************************************************************************
	public static void main(String args[]){
    	try{
			TestSolarSystem test = new TestSolarSystem();
			test.initFrame(); 		//System.out.println("initFrame OK");
			test.initUniverse(); 	//System.out.println("initUniverse OK");
	        test.displayCoord();
			test._canvas1.requestFocus();
        }
        catch(Exception e){
			e.printStackTrace();
            System.exit(0);
        }
	}// end main

	//****************************************************************************
	private void initFrame(){
    	// *** Set Look 'n' feel ***
		try{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		}catch (Exception e) { e.printStackTrace(); }
		// ***** Creation of GUI components *****
        String PATH_TO_IMAGES = ".." + FS + "data" + FS + "images" + FS;
		_frame = new JFrame("Test Systme solaire");
		_frame.addWindowListener(new FrameListener());
		_btnPlay = new JButton(new ImageIcon(PATH_TO_IMAGES + "play.gif"));
		_btnPlay.setEnabled(false);
		_btnPlay.addActionListener(new PlayListener());
		_btnPause = new JButton(new ImageIcon(PATH_TO_IMAGES + "pause.gif"));
		_btnPause.addActionListener(new PauseListener());
		_btnGo = new JButton("Go!");
		_btnGo.addActionListener(new GoListener());
		String[] strCombo = new String[SUN + 1];
        strCombo[0] = PLANET_NAMES[SUN];
		for(int i = 0; i < SUN ; i++) strCombo[i+1] = PLANET_NAMES[i]; // from Mercury to Pluto
		_cmbGo = new JComboBox(strCombo);
        _chkFromAbove = new JCheckBox("Vue de haut");
        _coordLabel = new JLabel("");
		_logArea = new JTextArea();
		_canvas1 = new Canvas3D (SimpleUniverse.getPreferredConfiguration()); //_canvas1 = new Canvas3D (null);
		_canvas2 = new Canvas3D (SimpleUniverse.getPreferredConfiguration()); //_canvas2 = new Canvas3D (null);
		_canvas2.setSize(200, 200);
		
		// ***** Layout *****
		Container contentPane = _frame.getContentPane();
//		contentPane.setLayout(new BorderLayout(10, 0));
		// CENTER
		contentPane.add(_canvas1, BorderLayout.CENTER);
		// EAST
		JPanel pEast = new JPanel();
		pEast.setLayout(new BoxLayout(pEast, BoxLayout.Y_AXIS));
		pEast.setBorder(BorderFactory.createEtchedBorder());
        // canvas2
		JPanel pCanvas2 = new JPanel();
		pCanvas2.setMaximumSize(pCanvas2.getSize());
		pCanvas2.add(_canvas2);
		pEast.add(pCanvas2);
        // pCommandGo : for chkFromAbove and panel Go
		JPanel pCommandGo = new JPanel();
		pCommandGo.setLayout(new BoxLayout(pCommandGo, BoxLayout.Y_AXIS));
//		pCommandGo.setBorder(BorderFactory.createEtchedBorder());
		_chkFromAbove.setAlignmentX(Component.RIGHT_ALIGNMENT);
        // pGo
		JPanel pGo = new JPanel(new FlowLayout(FlowLayout.LEFT));
		pGo.add(_cmbGo);
		pGo.add(_btnGo);
        pGo.setMaximumSize(pGo.getMinimumSize());
        pCommandGo.add(_chkFromAbove);
		pCommandGo.add(pGo);
        
		pEast.add(pCommandGo);
		pEast.add(javax.swing.Box.createVerticalGlue());
		
        if (DEBUG){ // play/pause
			JPanel pPlayPause = new JPanel(); // managed by FlowLayout
			pPlayPause.add(_btnPause);
			pPlayPause.add(_btnPlay);
			pEast.add(pPlayPause);
        }
		// logArea for tests
        if (DEBUG){
			JScrollPane areaScrollPane = new JScrollPane(_logArea);
			areaScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
			pEast.add(areaScrollPane);
        } // end logArea for tests
		contentPane.add(pEast, BorderLayout.EAST);
        //SOUTH
		contentPane.add(_coordLabel, BorderLayout.SOUTH);
		
		// ***** Packing *****
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		_frame.getRootPane().setPreferredSize(new Dimension(screen.width,screen.height - 55));
		_frame.pack();
		_frame.setVisible(true);
	}// end initFrame

	//****************************************************************************
	private void initUniverse(){
		_canvas = new Canvas3D(SimpleUniverse.getPreferredConfiguration()); // create a "ghost" Canvas3D
		_universe = new SimpleUniverse(_canvas);
		_universe.getViewingPlatform().getViewPlatform().setActivationRadius(getSemiAxis(PLUTO)*2);
		_universe.getViewer().getView().setBackClipDistance(getSemiAxis(PLUTO)*2);
		_universe.getViewer().getView().setFrontClipDistance(0.0);
		//_universe.getViewer().getView().setVisibilityPolicy(View.VISIBILITY_DRAW_ALL);
		//_universe.getViewingPlatform().setNominalViewingTransform();
		// Create one scene and two views, and add them to the universe
	    _universe.addBranchGraph(createSceneBG());
	    _universe.addBranchGraph(createViewBG(CANVAS1));
	    _universe.addBranchGraph(createViewBG(CANVAS2));
	}// end initUniverse

	//****************************************************************************
	/** Creates a view, associated with a canvas3D */
	private BranchGroup createViewBG(int whichCanvas){
		BranchGroup viewBG = new BranchGroup(); // returned value
		// choose on which canvas creating the view BG
		Canvas3D theCanvas = null;
		if (whichCanvas == CANVAS1) theCanvas = _canvas1;
		else if (whichCanvas == CANVAS2) theCanvas = _canvas2;
		
		// Initial position and direction of the viewer, depending on the canvas
		TransformGroup tg = new TransformGroup();
		Transform3D t3d = new Transform3D(); 
		if (whichCanvas == CANVAS1){
			t3d.lookAt(INITIAL_POS1, new Point3d(0, 0, 0), new Vector3d(0, 0, 1));
			t3d.invert();
		}
		else if (whichCanvas == CANVAS2){
			t3d.set(new Vector3f(0f, 0f, (float)(7 * PLANET_RADIUS[SUN]/SCALE_RADIUS)));
		}
		tg.setTransform(t3d);

		// Keyboard navigation
		if (whichCanvas == CANVAS1){
			tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
			tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
			tg.setCapability(TransformGroup.ALLOW_LOCAL_TO_VWORLD_READ);
            KeyBehavior kb = new KeyBehavior(tg);
			kb.setSchedulingBounds(BS);
			kb.setResetPosition(INITIAL_POS1);
            kb.setScaleFactor((float)(KM_PER_AU/SCALE_SEMIAXIS));
            kb.setCoordLabel(_coordLabel);
			tg.addChild(kb);
			// set the TG as instance variable for further modification (with btnGo)
			_viewerTG = tg;
		}

		// Build the tree
		ViewPlatform vp = new ViewPlatform();
		viewBG.addChild(tg);
		tg.addChild(vp);

		View view = new View();
		view.addCanvas3D(theCanvas);
        PhysicalBody pBody = new PhysicalBody();
        //pBody.setNominalEyeOffsetFromNominalScreen(0.0);
		view.setPhysicalBody(pBody);
		view.setPhysicalEnvironment(new PhysicalEnvironment());
		view.attachViewPlatform(vp);
		
		return viewBG;
	}// end createViewBG


	//****************************************************************************
	private BranchGroup createSceneBG(){
		BranchGroup root = new BranchGroup(); // returned value =root of the tree scene

		_planets = createPlanets(); // just creates the shapes
		_revolutionators = new RotationInterpolator[NB_PLANETS];
		Transform3D axisOfRotation; // used to reorient the graph with z axis up.

		// Creation of planet tree
		for (int iPlanet = 0; iPlanet <= SUN; iPlanet++){ // does not include the moon

			// Revolution of the planet = rotation around the sun
  			TransformGroup revolutionTG = new TransformGroup();
	      	revolutionTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE); // for play - pause

			if (iPlanet != SUN){
		      	revolutionTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE); // for play - pause
				Alpha revolutionAlpha = new Alpha(-1, (long)(PLANET_REVOLUTIONS[iPlanet]/SCALE_TIME_REVOLUTION));
				_revolutionators[iPlanet] = new RotationInterpolator(revolutionAlpha, revolutionTG);
				// set revolution axis as Z axis (default = y)
				axisOfRotation = _revolutionators[iPlanet].getAxisOfRotation(); 
				axisOfRotation.rotX(Math.PI/2);
				_revolutionators[iPlanet].setAxisOfRotation(axisOfRotation);
				_revolutionators[iPlanet].setSchedulingBounds(BS);
				revolutionTG.addChild(_revolutionators[iPlanet]);
			}

			// Initial translation to move the planet to its initial location
			TransformGroup initialTranslationTG = new TransformGroup();
			Transform3D initialTranslationT3D = new Transform3D();
			double posX = getSemiAxis(iPlanet) * Math.cos(PLANET_LAMBDAS[iPlanet]);
			double posY = getSemiAxis(iPlanet) * Math.sin(PLANET_LAMBDAS[iPlanet]);
			initialTranslationT3D.setTranslation(new Vector3f((float)(posX), (float)(posY), 0f));
			initialTranslationTG.setTransform(initialTranslationT3D);

			// Initial rotation to orient the axis
			TransformGroup initialRotationTG = new TransformGroup();
			Transform3D initialRotationT3D = new Transform3D();
			initialRotationT3D.rotY(Math.toRadians(PLANET_INCLINATION[iPlanet]));
			initialRotationTG.setTransform(initialRotationT3D);

			// Rotation of the planet around its axis
			TransformGroup spinTG = new TransformGroup();
			spinTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
			Alpha rotationAlpha = new Alpha(-1, (long)(PLANET_ROTATIONS[iPlanet]/SCALE_TIME_ROTATION));
			RotationInterpolator rotator = new RotationInterpolator(rotationAlpha, spinTG);
			// set rotation axis as Z axis (default = y)
			axisOfRotation = rotator.getAxisOfRotation(); 
			// don't know why must be different for the Sun, but it works...
			if (iPlanet != SUN) axisOfRotation.rotX(Math.PI/2);
			else axisOfRotation.rotX(- Math.PI/2);
			rotator.setAxisOfRotation(axisOfRotation);
			rotator.setSchedulingBounds(BS2);
			spinTG.addChild(rotator);
			
			// Rotation of the planet sphere of PI/2, to fit in the z axis
			TransformGroup planetSphereTG = new TransformGroup();
			Transform3D planetSphereT3D = new Transform3D();
			planetSphereT3D.rotX(Math.PI/2);
			planetSphereTG.setTransform(planetSphereT3D);

			// Build the scene tree
			root.addChild (revolutionTG);
			revolutionTG.addChild(initialTranslationTG);
			initialTranslationTG.addChild(initialRotationTG);
			initialRotationTG.addChild (spinTG);
			spinTG.addChild(planetSphereTG);
			planetSphereTG.addChild(_planets[iPlanet]);
			// add the rotation axis of the planet
			if (DRAW_AXIS[iPlanet]) spinTG.addChild(buildAxis(iPlanet));
			
			// Add the moon for the Earth
			if(iPlanet == EARTH){
			   	double radius = PLANET_RADIUS[MOON]; // km
        		double semiAxis = PLANET_SEMIAXIS[MOON] * KM_PER_AU; // km
                double revolution = PLANET_REVOLUTIONS[MOON] * 365.25;
                double rotation = revolution; // synchron
                initialTranslationTG.addChild(createSatellite(EARTH, radius, semiAxis, 
                				revolution, rotation, "MoonMap.jpg",
                                CHEAT_MOON_RADIUS, CHEAT_MOON_SEMIAXIS));
            }

			// Add the Sattelites for Jupiter
			if(iPlanet == JUPITER){
		        String[] fileNames = {"Jupiter_io.jpg", "Jupiter_Europa.jpg", "Jupiter_Ganymede.jpg", "Jupiter_Callisto.jpg"};
                for (int i=0; i<4; i++){
	                initialTranslationTG.addChild(createSatellite(JUPITER, JUPSAT_RADIUS[i],
                    			JUPSAT_SEMIAXIS[i]*PLANET_RADIUS[JUPITER], 
                				JUPSAT_REVOLUTIONS[i], JUPSAT_REVOLUTIONS[i], fileNames[i],
                                CHEAT_JUPSAT_RADIUS, CHEAT_JUPSAT_SEMIAXIS));
                }
            }

			// Add the rings for the Saturn
			if(iPlanet == SATURN){
				planetSphereTG.addChild(createSaturnRings());
            }

			if(iPlanet == PLUTO){
			   	double radius = 593; // km
        		double semiAxis = 19640; // km
                double revolution = PLANET_ROTATIONS[PLUTO]; // synchron
                double rotation = revolution; // synchron
                initialTranslationTG.addChild(createSatellite(PLUTO, radius, semiAxis, 
                				revolution, rotation, "Pluto_Charon.jpg",
                                CHEAT_CHARON_RADIUS, CHEAT_CHARON_SEMIAXIS));
            }
            
			// picking
            if (iPlanet == SATURN){
				revolutionTG.setCapability(TransformGroup.ENABLE_PICK_REPORTING);
            }
		
		}// end for (creation of planet tree)

		// Put point light inside the sun
		Light light = new PointLight (WHITE,
              						  new Point3f (0f, 0f, 0f),		   		// position
                  					  new Point3f (0.00001f, 0f, 0f));  	// attenuation
		light.setInfluencingBounds(BS);

		// picking
		PickTranslateBehavior pick = new  PickTranslateBehavior(root, _canvas1, BS);
		PickZoomBehavior pick2 = new  PickZoomBehavior(root, _canvas1, BS);
		PickRotateBehavior pick3 = new  PickRotateBehavior(root, _canvas1, BS);
		root.addChild(pick);
		root.addChild(pick2);
		root.addChild(pick3);

		// Background
		Background bg = new Background(new Color3f(Color.black));
	    bg.setApplicationBounds(new BoundingBox());
		
		// Complete the scene tree
		root.addChild (bg);
		root.addChild (light);
		root.compile();
		return root;

	}// end createSceneBG

	//****************************************************************************
	/** Auxiliary method used by createSceneBG */
	private Sphere[] createPlanets(){

		Sphere[] planets = new Sphere[NB_PLANETS];
		// files where the textures are stored
		String[] filenames = {"MercuryMap.jpg", "VenusMap.jpg", "EarthMap.jpg", "MarsMap.jpg", "JupiterMap.jpg", "SaturnMap.jpg",
							  "UranusMap.jpg", "NeptuneMap.jpg", "PlutonMap.jpg", "SunMap2.jpg", "MoonMap.jpg"};
		// variables used several times
		String filename;
		Texture tex;
		Appearance planetAppearance;
		TextureAttributes texAttr;
		// Creation of planets
		for (int iPlanet = MERCURY; iPlanet <= MOON; iPlanet++){
			planetAppearance = new Appearance();
			filename = TEXTURE_DIR + filenames[iPlanet];
			tex = new TextureLoader(filename, _canvas1).getTexture();
			planetAppearance.setTexture(tex);
			texAttr = new TextureAttributes();
			texAttr.setTextureMode(TextureAttributes.MODULATE);
			planetAppearance.setTextureAttributes(texAttr);		
			planetAppearance.setMaterial(new Material());
            if (iPlanet == SUN){
				planets[iPlanet] = new Sphere((float)(PLANET_RADIUS[iPlanet]/SCALE_RADIUS*CHEAT_RADIUS[iPlanet]), 
												Primitive.GENERATE_TEXTURE_COORDS | Primitive.GENERATE_NORMALS_INWARD, planetAppearance);
			}
			else{
				planets[iPlanet] = new Sphere((float)(PLANET_RADIUS[iPlanet]/SCALE_RADIUS*CHEAT_RADIUS[iPlanet]), 
												Primitive.GENERATE_TEXTURE_COORDS | Primitive.GENERATE_NORMALS, planetAppearance);
			}
			planets[iPlanet].setCapability(Node.ALLOW_LOCAL_TO_VWORLD_READ);
		}// end for

		return planets;
	}// end createPlanets

	//****************************************************************************
	/** Creation of Saturn rings */
	private Cylinder createSaturnRings(){
		Appearance RingsAppearance = new Appearance();
		Texture texRings = new TextureLoader(TEXTURE_DIR + "SaturnRings.jpg", _canvas1).getTexture();
		RingsAppearance.setTexture(texRings);
		TextureAttributes texAttrRings = new TextureAttributes();
		texAttrRings.setTextureMode(TextureAttributes.MODULATE);
		RingsAppearance.setTextureAttributes(texAttrRings);
		Material matRings = new Material();
		RingsAppearance.setMaterial(matRings);
    	// radius of the most external ring, in Saturn radii
    	double ringSize = 7.95;
        double radius = PLANET_RADIUS[SATURN] / SCALE_RADIUS * ringSize;
		return new Cylinder(1.5f,0.01f,Cylinder.GENERATE_TEXTURE_COORDS | Cylinder.GENERATE_NORMALS ,RingsAppearance);    	
    }// end createSaturnRings

	//****************************************************************************
	/** Creation of a satellite
    @returns a part of the scene containing the satellite and its movements
	@param iPlanet index of the planet concerned by the satellite
    @param radius radius of the satellite in km
    @param semiAxis semi-major axis of the satellite in planetary radii
    @param rotation rotation period in terrestrial days
    @param revolution revolution period in terrestrial days
    @param fileName Name of the file containing the texture (must be in TEXTURE_DIR)
    */
    private TransformGroup createSatellite(int iPlanet, double radius, double semiAxis, 
    										double rotation, double revolution,
                                            String fileName,
                                            double cheatOnRadius, double cheatOnSemiAxis){
	
		// Revolution of the satellite around its planet
		TransformGroup revolutionTG = new TransformGroup();
  		revolutionTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		Alpha revolutionAlpha = new Alpha(-1, (long)(rotation/SCALE_TIME_ROTATION));
		RotationInterpolator revolutionator = new RotationInterpolator(revolutionAlpha, revolutionTG);
		// set rotaion axis as Z axis (default = y)
		Transform3D axisOfRotation = revolutionator.getAxisOfRotation(); 
		axisOfRotation.rotX(Math.PI/2);
		revolutionator.setAxisOfRotation(axisOfRotation);
		revolutionator.setSchedulingBounds(BS); // could be lowered
		revolutionTG.addChild(revolutionator);

		// Move the satellite to its initial pos
        // Warning : the scale factor is SCALE_RADIUS
        // because the visual rendering is related to the planet sphere, then at radius scale.
		TransformGroup initialTranslationTG = new TransformGroup();
		Transform3D initialT3D = new Transform3D();
		double posX = semiAxis/SCALE_RADIUS * CHEAT_RADIUS[iPlanet] / cheatOnSemiAxis;
		double posY = 0;
		initialT3D.setTranslation(new Vector3f((float)(posX), (float)(posY), 0f));
		initialTranslationTG.setTransform(initialT3D);
		initialTranslationTG.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		initialTranslationTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        
		// Rotation of the satellite around its axis
		TransformGroup spinTG = new TransformGroup();
		spinTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        // WARNING : scaling with SCALE_TIME_ROTATION because revolution expressed in days
		Alpha rotationAlpha = new Alpha(-1, (long)(revolution/SCALE_TIME_ROTATION));
		RotationInterpolator rotator = new RotationInterpolator(rotationAlpha, spinTG);
		// set rotion axis as Z axis (default = y)
		axisOfRotation = rotator.getAxisOfRotation(); 
		axisOfRotation.rotX(Math.PI/2);
		rotator.setAxisOfRotation(axisOfRotation);
		rotator.setSchedulingBounds(BS);
		spinTG.addChild(rotator);

		Appearance app = new Appearance();
		fileName = TEXTURE_DIR + fileName;
		Texture tex = new TextureLoader(fileName, _canvas1).getTexture();
		app.setTexture(tex);
		TextureAttributes texAttr = new TextureAttributes();
		texAttr.setTextureMode(TextureAttributes.MODULATE);
		app.setTextureAttributes(texAttr);		
		app.setMaterial(new Material());
        // Build the sphere representing the satellite
		Sphere satellite = new Sphere((float)(radius / SCALE_RADIUS * CHEAT_RADIUS[iPlanet] * cheatOnRadius),
										Primitive.GENERATE_TEXTURE_COORDS | Primitive.GENERATE_NORMALS, 
										app);

		// Rotation of Satellite sphere of PI/2, to fit in the z axis
		TransformGroup sphereTG = new TransformGroup();
		Transform3D sphereT3D = new Transform3D();
		sphereT3D.rotX(Math.PI/2);
		sphereTG.setTransform(sphereT3D);

		// Create the local branch group
		revolutionTG.addChild(initialTranslationTG);
		initialTranslationTG.addChild(spinTG);
		spinTG.addChild(sphereTG);
		sphereTG.addChild(satellite);
        
        return revolutionTG;
    } // end createSatellite

	//****************************************************************************
	/** Auxiliary method used to create planet rotation axis */
	private static Shape3D buildAxis(int iPlanet){
		float axisLen = (float)(PLANET_RADIUS[iPlanet]*CHEAT_RADIUS[iPlanet]/SCALE_RADIUS * 2);
		float arrowLen = axisLen / 18f;
		LineArray axisLine = new LineArray(6, LineArray.COORDINATES | LineArray.COLOR_3);
		// main line
		axisLine.setCoordinate(0, new Point3f(0.0f, 0.0f, -axisLen));
		axisLine.setCoordinate(1, new Point3f(0.0f, 0.0f, axisLen));
		// first part of arrow
		axisLine.setCoordinate(2, new Point3f(0.0f, 0.0f, axisLen));
		axisLine.setCoordinate(3, new Point3f(arrowLen / 2f, 0.0f, axisLen - arrowLen));
		// second part os arrow
		axisLine.setCoordinate(4, new Point3f(0.0f, 0.0f, axisLen));
		axisLine.setCoordinate(5, new Point3f(- arrowLen / 2f, 0.0f, axisLen - arrowLen));
		for (int i = 0; i < 6; i++)
		  	axisLine.setColor(i, WHITE);
		return new Shape3D(axisLine);
	}// end buildAxis
	
	//****************************************************************************
	/** Returns the semi-major axis of a planet, in units of the virtual world's coordinate system. */
	private static float getSemiAxis(int iPlanet){
		return (float)(PLANET_SEMIAXIS[iPlanet]*KM_PER_AU/SCALE_SEMIAXIS);
		//return (float)(Math.log(PLANET_SEMIAXIS[iPlanet]+ 10));
	}// end getSemiAxis
	
	//************************* displayCoord() *******************************************
    private static NumberFormat NF; // to display the coords
    static{
  	  NF = NumberFormat.getInstance();
      NF.setMaximumFractionDigits(3);
      NF.setMinimumFractionDigits(3);
    };
    private static final float coordDisplayScaleFactor = (float)(KM_PER_AU/SCALE_SEMIAXIS);
    private static final String STR1 = "coords : (";
    private static final String STR2 = " ";
    private static final String STR3 = ") - dist : ";
    private static final String STR4 = " a.u. ";
    /** 
    Displays the coords of the viewer in _coordLabel
    */
	private void displayCoord(){
    	Transform3D t3D = new Transform3D();
	    _viewerTG.getTransform(t3D);
		Vector3f v = new Vector3f();
		t3D.get(v); // vector 'v' recieves the translational component of 'transform3D'
	    _coordLabel.setText(STR1 + NF.format(v.x/coordDisplayScaleFactor) + STR2 
        						 + NF.format(v.y/coordDisplayScaleFactor) + STR2 
                                 + NF.format(v.z/coordDisplayScaleFactor) + STR3
                                 + NF.format(v.length()/coordDisplayScaleFactor) + STR4);
    }// end displayCoord()

	//****************************************************************************
	private static BranchGroup LoadFile(String sFileName){
		try{
			ObjectFile f = new ObjectFile();
			Scene s = f.load (sFileName);
			return s.getSceneGroup();
		}
		catch (Exception e){
			return null;
		}
	}// end LoadFile

	//****************************************************************************
	private static void loadConfig (){
    	String PATH_TO_CONFIG = ".." + FS + "data" + FS + "config" + FS;
        // Keys of the config files
        String WHICH_CONFIG_KEY 			= "whichConfig";
		String SCALE_RADIUS_KEY 			= "SCALE_RADIUS";
		String SCALE_SEMIAXIS_KEY 			= "SCALE_SEMIAXIS";
		String ROTATION_SCALE_KEY 			= "ROTATION_SCALE";
		String REVOLUTION_SCALE_KEY 		= "REVOLUTION_SCALE";
		String DRAW_AXIS_KEY 				= "DRAW_AXIS";
        String CHEAT_RADIUS_KEY				= "CHEAT_RADIUS";
		String CHEAT_MOON_SEMIAXIS_KEY		= "CHEAT_MOON_SEMIAXIS";
		String CHEAT_MOON_RADIUS_KEY		= "CHEAT_MOON_RADIUS";
		String CHEAT_JUPSAT_RADIUS_KEY 		= "CHEAT_JUPSAT_RADIUS";
		String CHEAT_JUPSAT_SEMIAXIS_KEY	= "CHEAT_JUPSAT_SEMIAXIS";
		String CHEAT_CHARON_RADIUS_KEY 		= "CHEAT_CHARON_RADIUS";
		String CHEAT_CHARON_SEMIAXIS_KEY	= "CHEAT_CHARON_SEMIAXIS";
		// intermediate variables
        double rotation_scale = 1;
		double revolution_scale = 1;

		try{
        	// Read in 'whichConf.txt' which config file must be used.
			Properties p = new Properties();
            p.load(new FileInputStream(new File(PATH_TO_CONFIG + "whichConfig.txt")));
            String whichConfig = p.getProperty(WHICH_CONFIG_KEY);
			// Read the config infos in the selected config file.
            p.load(new FileInputStream(new File(PATH_TO_CONFIG + whichConfig)));
            SCALE_RADIUS 		= Double.parseDouble(p.getProperty(SCALE_RADIUS_KEY));
            SCALE_SEMIAXIS 		= Double.parseDouble(p.getProperty(SCALE_SEMIAXIS_KEY));
            rotation_scale 		= Double.parseDouble(p.getProperty(ROTATION_SCALE_KEY));
            revolution_scale 	= Double.parseDouble(p.getProperty(REVOLUTION_SCALE_KEY));
            // draw axis
            for(int i = 0; i < DRAW_AXIS.length; i++) DRAW_AXIS[i] = false;
            String[] strArray = stringToStringArray(p.getProperty(DRAW_AXIS_KEY));
            for(int i = 0; i < strArray.length; i++){
            	DRAW_AXIS[getIntConstant(strArray[i], "TestSolarSystem")] = true;
            }
            // Cheat factors
            String[] cheatRadii = stringToStringArray(p.getProperty(CHEAT_RADIUS_KEY));
            for(int i = 0; i < cheatRadii.length; i++){
            	CHEAT_RADIUS[i] = Double.parseDouble(cheatRadii[i]);
            }
            CHEAT_MOON_SEMIAXIS 	= Double.parseDouble(p.getProperty(CHEAT_MOON_SEMIAXIS_KEY));
            CHEAT_MOON_RADIUS	 	= Double.parseDouble(p.getProperty(CHEAT_MOON_RADIUS_KEY));
			CHEAT_JUPSAT_RADIUS	 	= Double.parseDouble(p.getProperty(CHEAT_JUPSAT_RADIUS_KEY));
			CHEAT_JUPSAT_SEMIAXIS 	= Double.parseDouble(p.getProperty(CHEAT_JUPSAT_SEMIAXIS_KEY));
			CHEAT_CHARON_RADIUS	 	= Double.parseDouble(p.getProperty(CHEAT_CHARON_RADIUS_KEY));
			CHEAT_CHARON_SEMIAXIS 	= Double.parseDouble(p.getProperty(CHEAT_CHARON_SEMIAXIS_KEY));
		}
		catch (Exception e){
        	e.printStackTrace();
			System.out.println("Error while loading config files");
			System.out.println("Default values taken");
			SCALE_RADIUS 			= 1E5;
			SCALE_SEMIAXIS 			= 5E6;
            for(int i = 0; i < DRAW_AXIS.length; i++) DRAW_AXIS[i] = true;
            for(int i = 0; i < CHEAT_RADIUS.length; i++) CHEAT_RADIUS[i] = 1;
			CHEAT_MOON_SEMIAXIS		= 1;
			CHEAT_MOON_RADIUS		= 1;
			CHEAT_JUPSAT_RADIUS	 	= 1;
			CHEAT_JUPSAT_SEMIAXIS 	= 1;
			CHEAT_CHARON_RADIUS	 	= 1;
			CHEAT_CHARON_SEMIAXIS 	= 1;
		}
        finally{
			SCALE_TIME_ROTATION = 1 / (rotation_scale * 1000.0);
			SCALE_TIME_REVOLUTION = 1 / (revolution_scale * 365.25 * 1000.0);
        }
	}// end loadConfig

  //******************************************************************
  // Used by loadConfig - taken from an other package
  /** Convenient method to get a <CODE>static final int</CODE> field of a class, using reflection.
  @throws RuntimeException If the property can't be correctly retrieved.
  */
	public static int getIntConstant(String value, String className){
    try{
      Class theClass = Class.forName(className);
      Field field = theClass.getDeclaredField(value);
      return field.getInt(null);
    }
    catch(Exception e){
      throw new RuntimeException(e.toString());
    }
  }// end getIntConstant
  
  //******************* stringToStringArray(String s) **************	
  // Used by loadConfig - taken from an other package
  /** Convenient method to call {@link #stringToStringArray(String, char)}, 
  using a coma (',') as separator. */
	public static String[] stringToStringArray(String s){ return stringToStringArray(s, ','); }

	//******************* stringToStringArray(String s, char separator) **************	
    // Used by loadConfig - taken from an other package
	/**	Transforms a <CODE>String</CODE> in an array of <CODE>String</CODE>s.
	<BR/><B>Example</B> : if <CODE>s = "arg1, arg2,arg3 "</CODE> and <CODE>separator = ','</CODE>,
 	<CODE>stringToStringArray</CODE> returns : 
	<TABLE BORDER><TR>
	<TD><FONT SIZE="-1">arg1</FONT></TD>
	<TD><FONT SIZE="-1">arg2</FONT></TD>
	<TD><FONT SIZE="-1">arg3</FONT></TD>
	</TR></TABLE>
  Points to consider :
	<LI>Leading and trailing spaces are removed (see <CODE>arg2</CODE> and <CODE>arg3</CODE>).</LI>
	<LI>Separators at the begining or at the end of <CODE>s</CODE> lead to creation 
  of empty Strings in the resulting array.</LI>
	<LI>The same thing happens if two separators are side by side : 
	<BR/>
	if <CODE>s=",arg1,, arg2,arg3"</CODE>, <CODE>stringToStringArray</CODE> returns : 
	<TABLE BORDER><TR>
	<TD><I><FONT SIZE="-1">(empty String)</FONT></I></TD>
	<TD><FONT SIZE="-1">arg1</FONT></TD>
	<TD><I><FONT SIZE="-1">(empty String)</FONT></I></TD>
	<TD><FONT SIZE="-1">arg2</FONT></TD>
	<TD><FONT SIZE="-1">arg3</FONT></TD>
	</TR></TABLE>
  </LI>

	@param s String to analyze.
	@param separator Character used as separator to analyze <CODE>s</CODE>.
	@return An array of Strings resulting from the analysis.
	*****************************************************************/
	public static String[] stringToStringArray(String s, char separator){
		//System.out.println("s = " + s);
		int curPos=0, i=0;
		int sepPos=s.indexOf(separator, curPos);
		
		//1er passage, dtermination de la longueur du tableau rultat res[]
		while(sepPos!=-1){
			curPos=sepPos+1;
			sepPos=s.indexOf(separator, curPos);
			i++;
		}
		//2me passage, remplissage de res[]
		//System.out.println("2eme passage");
		String[] res = new String[i+1];
		curPos=0;
		i=0;
		sepPos=s.indexOf(separator, curPos);
		while(sepPos!=-1){
			res[i]=s.substring(curPos, sepPos).trim();
			//System.out.println("res[" + i + "] = " + res[i] + " ; sepPos = " + sepPos);
			curPos=sepPos+1;
			sepPos=s.indexOf(separator, curPos);
			i++;
		}
			res[i]=s.substring(curPos, s.length()).trim();
			//System.out.println("res[" + i + "] = " + res[i] + " ; sepPos = " + sepPos);
    return res;
  }//end stringToStringArray

	//=================================================================================
	//                            INTERNAL CLASSES
	//=================================================================================
	
	//****************************************************************************
	class FrameListener extends WindowAdapter{
		public void windowClosing (WindowEvent e){
			System.exit(0);
		}
	}// end class FrameListener
	
	
	//****************************************************************************
	class PlayListener implements ActionListener{
		public void actionPerformed(ActionEvent e){
			_btnPlay.setEnabled(false);
			_btnPause.setEnabled(true);
			for (int i = MERCURY; i == MOON; i++){
				_revolutionators[i].setEnable(true);
			}
		}
	}// end class PlayListener
	
	//****************************************************************************
	class PauseListener implements ActionListener{
		public void actionPerformed(ActionEvent e){
			_btnPlay.setEnabled(true);
			_btnPause.setEnabled(false);
			for (int i = MERCURY; i == MOON; i++){
				_revolutionators[i].setEnable(false);
			}
		}
	}// end class PauseListener
	
	//****************************************************************************
	/** Permits to move the user to a given planet */
	class GoListener implements ActionListener{
		public void actionPerformed(ActionEvent e){
			int iPlanet = _cmbGo.getSelectedIndex();
            // transform to the right index (the sun appears first in the list)
            if (iPlanet == 0) iPlanet = SUN;
            else iPlanet --;

			// Get the position of the planet
			Transform3D t = new Transform3D();
			_planets[iPlanet].getLocalToVworld(t); // get the T3D of the planet in virtual world coord
			Vector3f v = new Vector3f();
			t.get(v); // vector 'v' recieves the translational component of 't' (the planet pos)
			Point3d planetPos = new Point3d(v);
			
			// The eye is put at 'n' planet radii from the planet.
			float n = 0;
            switch(iPlanet){
            	case SUN:
                	n = 400; break;
            	case MERCURY:
                	n = 1800; break;
            	case VENUS: case EARTH: case MARS:
                	n = 1400; break;
	            case JUPITER: case SATURN: 
                	if(_chkFromAbove.isSelected()) n = 1000;
                    else n = 600;
                    break;
                case URANUS: case NEPTUNE:
                	n = 400; break;
            	case PLUTO:
                	n = 4000; break;
            }

			// Calculate the position of the viewer	
			Point3d viewerPos;
            Vector3d upVector;
			float deltaR = (float)(n * _planets[iPlanet].getRadius() * SCALE_RADIUS);
            if (_chkFromAbove.isSelected()){
				viewerPos = new Point3d(v.x, v.y, v.z + (double)deltaR/SCALE_SEMIAXIS);
            	if (iPlanet == SUN){
	                upVector = new Vector3d(0,1,0);
                }
                else{
	                upVector = new Vector3d(v);
	                upVector.normalize();
                }
            }
            else{ // put the viewer between the planet and the sun
				// compute the values in units of real world(km)
            	if (iPlanet == SUN){
					viewerPos = INITIAL_POS1;
                }
                else{
					float rho = (float)(v.length() * SCALE_SEMIAXIS);
					rho = (float)((rho - deltaR) / SCALE_SEMIAXIS);
					double factor = rho/v.length(); 
					v.scale((float)factor);
					viewerPos = new Point3d(v);
                 }
                upVector = new Vector3d(0, 0, 1);
            }

			Transform3D tFinal = new Transform3D();
			tFinal.lookAt(viewerPos, planetPos, upVector);
			tFinal.invert();
			//tFinal.setTranslation(new Vector3d(viewerPos)); // normally useless ?
     		_viewerTG.setTransform(tFinal);
            displayCoord();
			_canvas1.requestFocus();
		}
	}// end class GoListener
	
}// end class TestSolarSystem

