//*********************************************************************************
// class TestStars
// Software released under the General Public License (version 2 or later), available at
// http://www.gnu.org/copyleft/gpl.html
//*********************************************************************************
import sky.*;

import java.util.*;
import java.io.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.*;

import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.universe.SimpleUniverse;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.behaviors.mouse.*;
import com.sun.j3d.utils.behaviors.keyboard.*;
import com.sun.j3d.utils.picking.*;
import com.sun.j3d.utils.picking.behaviors.*;
import com.sun.j3d.loaders.*;
import com.sun.j3d.loaders.objectfile.*;
/*
WARNING : bug related with 'NameBehaviour.java' : a click does not permit to get the wanted star name, but several names.
*/
public class TestStars{

  //=================================================================================
  //                            INSTANCE VARIABLES
  //=================================================================================
	// Graphical fields
	private static SimpleUniverse	 	_universe;
	private static JFrame 					_frame;
	private static Canvas3D 				_canvas;
	private static JTextField				_statusBar;
	
  //=================================================================================
  //                            CONSTANTS
  //=================================================================================

	// Star data
	private final static int 				NB_STARS = 109;
	private final static double[][] STAR_COORDS = NearestStarsData.coords;
	private final static String[][] STAR_NAMETYPE = NearestStarsData.nameType;
	private static final double 		SCALE_FACTOR = 5.0;
	private static final float 			STAR_RADIUS = 1E-2f;

	// Colors
	private static final Color3f black = new Color3f(Color.black);
	private static final Color3f blue = new Color3f(Color.blue);
	private static final Color3f white = new Color3f(Color.white);
	private static final Color3f yellow = new Color3f(Color.yellow);
	private static final Color3f orange = new Color3f(Color.orange);
	private static final Color3f red = new Color3f(Color.red);

  //=================================================================================
  //                                      METHODS
  //=================================================================================

	//****************************************************************************
	public static void main(String args[]){
		try{
			initFrame();
			initScene();
		}
		catch (Throwable e){
	    try{
	      System.setErr(new PrintStream(new FileOutputStream(new File("errors.txt"))));
		    System.err.println(e.toString());
	  		//e.printStackTrace();
	      //System.exit(0);
	    }
	    catch(Exception e2){
		  	System.out.println("*** Debug.traceError - Failed to trace an exception ***");
	    }
		}
	}

	//****************************************************************************
	private static void initFrame(){
		// creation of GUI components
		_frame = new JFrame("Tests 3D");
		_statusBar = new JTextField("Status bar");
		_statusBar.setEditable(false);

		// Canvas3D and universe creation
	  _canvas = new Canvas3D (SimpleUniverse.getPreferredConfiguration());
		_canvas.setVisible(true);
	  _universe = new SimpleUniverse(_canvas);
		
		// Layout and listeners
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		_frame.getRootPane().setPreferredSize(new Dimension(screen.width,screen.height));
		_frame.getContentPane().add(_canvas, BorderLayout.CENTER);
		_frame.getContentPane().add(_statusBar, BorderLayout.SOUTH);
		_frame.addWindowListener(new FrameListener());
		_frame.pack();
		_frame.setVisible(true);
	}

	//****************************************************************************
	private static void initScene(){
	  // Cration de la scne 3D  visualiser
		BranchGroup scene = createSceneTree();
		// Rattachement de la scne 3D  l'univers
		_universe.addBranchGraph(scene);
		// Positionnement pour avoir une vue correcte sur la scne 3D 
		// (permet de voir une scne 3D contenue dans un cube d'1 unit  
		// de ct et centr sur le centre du repre).
		_universe.getViewingPlatform().setNominalViewingTransform();
	}

	//****************************************************************************
	private static BranchGroup createSceneTree(){
		// root of the tree scene
		BranchGroup root = new BranchGroup();
		
		// Background
		Background bg = new Background(new Color3f(Color.black));
    bg.setApplicationBounds(new BoundingBox());
		
		// Materials (to make the stars glow) depending on spectral types (O, B, A, F, G, K, M)
		Material matOBA = new Material();
		matOBA.setEmissiveColor(blue);
		matOBA.setShininess(128.0f);
		matOBA.setLightingEnable(true);
		Material matF = new Material();
		matF.setEmissiveColor(white);
		matF.setShininess(128.0f);
		matF.setLightingEnable(true);
		Material matG = new Material();
		matG.setEmissiveColor(yellow);
		matG.setShininess(128.0f);
		matG.setLightingEnable(true);
		Material matK = new Material();
		matK.setEmissiveColor(orange);
		matK.setShininess(128.0f);
		matK.setLightingEnable(true);
		Material matM = new Material();
		matM.setEmissiveColor(red);
		matM.setShininess(128.0f);
		matM.setLightingEnable(true);
		
		// Appearences for the different materials
		Appearance appOBA = new Appearance();
		appOBA.setMaterial(matOBA);
		Appearance appF = new Appearance();
		appF.setMaterial(matF);
		Appearance appG = new Appearance();
		appG.setMaterial(matG);
		Appearance appK = new Appearance();
		appK.setMaterial(matK);
		Appearance appM = new Appearance();
		appM.setMaterial(matM);

		// Create stars
		String type;
		for (int i = 0; i < NB_STARS; i++){
			// Build the star
			Sphere star = new Sphere(STAR_RADIUS);

			// Set the appearence depending on the star's spectral type
			type = STAR_NAMETYPE[i][1];
			if(type.substring(0,1).toLowerCase().equals("f"))
				star.setAppearance(appF);
			else if(type.substring(0,1).toLowerCase().equals("g"))
				star.setAppearance(appG);
			else if(type.substring(0,1).toLowerCase().equals("k"))
				star.setAppearance(appK);
			else if(type.substring(0,1).toLowerCase().equals("m"))
				star.setAppearance(appM);
			else if(type.substring(0,1).toLowerCase().equals("o") ||
				  		type.substring(0,1).toLowerCase().equals("b") ||
				  		type.substring(0,1).toLowerCase().equals("a")
						 )
				star.setAppearance(appOBA);
			else // default value : white star
				star.setAppearance(appF);

			// Star transform group
			TransformGroup initialTranslationTG = new TransformGroup();
			Vector3f v3f = new Vector3f((float)(STAR_COORDS[i][0]/SCALE_FACTOR),
																	(float)(STAR_COORDS[i][1]/SCALE_FACTOR),
																	(float)(STAR_COORDS[i][2]/SCALE_FACTOR));
			Transform3D translateT3D = new Transform3D();
			translateT3D.setTranslation(v3f);
			initialTranslationTG.setTransform(translateT3D);
			
			// Create the star branch group (obliged to have a BG for NameBehaviour).
			BranchGroup starBG = new BranchGroup();
			// Add the transformGroup to root
			root.addChild(starBG);
			starBG.addChild(initialTranslationTG);
			initialTranslationTG.addChild(star);

			// handle response to mouse events
			BoundingSphere bs = new BoundingSphere(); // in the local frame of initialTranslationTG
			bs.setRadius((double)STAR_RADIUS);
			NameBehaviour nb = new NameBehaviour(starBG, _canvas, _statusBar, STAR_RADIUS); 
			nb.setPickable(true);
			nb.setUserData(STAR_NAMETYPE[i][0]); // associate the behaviour with star name
			nb.setSchedulingBounds(bs);
			nb.setCapability(Node.ENABLE_PICK_REPORTING);
			nb.setCapability(Node.ALLOW_PICKABLE_READ);
			nb.setCapability(Node.ALLOW_BOUNDS_READ);
			starBG.addChild(nb);

		}// end for (creation of stars)

		// Create sun at center of frame
		Sphere sun = new Sphere(1E-2f);
		sun.setAppearance(appG);
		root.addChild(sun);

  	// keybord navigation
		KeyNavigatorBehavior mkey = new KeyNavigatorBehavior(_universe.getViewingPlatform().getViewPlatformTransform());
		mkey.setSchedulingBounds(new BoundingSphere(new Point3d(0,0,0),1000.0)); 
		
		// light
		Color3f lightColor = new Color3f (1, 1, 0.5f); // light yellow
		Light light = new PointLight (lightColor,
              		new Point3f (0, 1, 1),   				// position
                  new Point3f (1f, 0, 0));  			// attenuation
		light.setInfluencingBounds (new BoundingSphere (new Point3d (), 100.0));

		// picking
		PickTranslateBehavior pick = new  PickTranslateBehavior(root, _canvas, new BoundingSphere());
		PickZoomBehavior pick2 = new  PickZoomBehavior(root, _canvas, new BoundingSphere());
		PickRotateBehavior pick3 = new  PickRotateBehavior(root, _canvas, new BoundingSphere());
		root.addChild(pick);
		root.addChild(pick2);
		root.addChild(pick3);
		
		// scene tree
		root.addChild (bg);
		root.addChild (light);
		root.addChild(mkey);
		
		root.compile();
		
		return root;
	}

	//****************************************************************************
	private static BranchGroup LoadFile(String sFileName){
		try{
		ObjectFile f = new ObjectFile();
		Scene s = f.load (sFileName);
		return s.getSceneGroup();
		}
		catch (Exception e){
			return null;
		}
	}// end loadFile

}// end class TestStars

//=================================================================================
//                            INTERNAL CLASSES
//=================================================================================
class FrameListener extends WindowAdapter{
	public void windowClosing (WindowEvent e){
		System.exit(0);
	}
}// end class FrameListener

