//********************************************************************
// class sky.NearestStarsData
// Software released under the General Public License (version 2 or later), available at
// http://www.gnu.org/copyleft/gpl.html
//********************************************************************
package sky;

/********************************************************************
Class containing data of the 109 nearest stars.
<LI><CODE>nameType</CODE> contains names and spectral types.</LI>
<LI><CODE>coords</CODE> contains cartesian coordinates in FK5 equatorial frame ( source : Hipparcos).</LI>
********************************************************************/
public abstract class NearestStarsData{

  /** Each line contains the name and the spectral type. */
  public static final String[][] nameType = {
    {"Proxima", "M5.5"},
    {"Alpha Centauri A", "G2"},
    {"Alpha Centauri B", "K0"},
    {"Barnard's Star", "M5"},
    {"Wolf 359", "M6"},
    {"Lalande 21185", "M2"},
    {"Sirius A", "A1"},
    {"Sirius B", "DA2"},
    {"L 726-8 A", "M5.5"},
    {"L 726-8 B", "M5.5"},
    {"Ross 154", "M4.5"},
    {"Ross 248", "M6"},
    {"Epsilon Eridani", "K2"},
    {"Lacaille 9352", "M2"},
    {"Ross 128", "M4.5"},
    {"L 789-6 A", "M5.5"},
    {"L 789-6 B", "M5"},
    {"L 789-6 C", "m"},
    {"Procyon A", "F5"},
    {"Procyon B", "DA"},
    {"61 Cygni A", "K5"},
    {"61 Cygni B", "K7"},
    {"Struve 2398 A", "M4"},
    {"Struve 2398 B", "M5"},
    {"Groombridge 34 A", "M2"},
    {"Groombridge 34 B", "M6"},
    {"G51-15", "M6.5"},
    {"Epsilon Indi", "K5"},
    {"Tau Ceti", "G8"},
    {"L 372-58", "M5.5"},
    {"L 725-32", "M5"},
    {"Luyten's Star", "M3.5"},
    {"Kapteyn's Star", "M0"},
    {"Lacaille 8760", "M0"},
    {"Kruger 60 A", "M2"},
    {"Kruger 60 B", "M6"},
    {"Ross 614 A", "M4.5"},
    {"Ross 614 B", "m"},
    {"Wolf 1061", "M3.5"},
    {"Wolf 424 A", "M5.5"},
    {"Wolf 424 B", "M7"},
    {"CD-37 15492", "M4"},
    {"van Maanen's Star", "DZ7"},
    {"L 1159-16", "M8"},
    {"L 143-23", "M5.5"},
    {"LP 731-58", "M6.5"},
    {"BD+68 946", "M3.5"},
    {"CD-46 11540", "M3"},
    {"L 145-141", "DQ6"},
    {"G158-27", "M5.5"},
    {"Ross 780", "M5"},
    {"G208-44 A", "M5.5"},
    {"G208-44 B", "M6"},
    {"G208-44 C", "m"},
    {"Lalande 21258 A", "M2"},
    {"Lalande 21258 B", "M6"},
    {"Groombridge 1618", "K2"},
    {"BD+20 2465", "M4.5"},
    {"L 354-89", "M1"},
    {"LP 944-20", "L"},
    {"CD-44 11909", "M3.5"},
    {"Omicron^2 Eridani A", "K1"},
    {"Omicron^2 Eridani B", "DA4"},
    {"Omicron^2 Eridani C", "M4.5"},
    {"BD+43 4305", "M4.5"},
    {"70 Ophiuchi A", "K0"},
    {"70 Ophiuchi B", "K5"},
    {"Altair", "A7"},
    {"G9-38 A", "M5.5"},
    {"G9-38 B", "M5.5"},
    {"L 722-22 A", "M4"},
    {"L 722-22 B", "m"},
    {"G99-49", "M4"},
    {"G254-29", "M4"},
    {"Lalande 25372", "M4"},
    {"LP 816-60", "m"},
    {"Stein 2051 A", "M4"},
    {"Stein 2051 B", "DC5"},
    {"Wolf 294", "M4"},
    {"Wolf 1453", "M1.5"},
    {"L 347-14", "M4.5"},
    {"Wolf 630 A", "M3"},
    {"Wolf 630 B", "M4"},
    {"Wolf 630 C", "M7"},
    {"Sigma Draconis", "K0"},
    {"L 668-21 A", "M1"},
    {"L 668-21 B", "T"},
    {"Ross 47", "M4"},
    {"L 205-128", "M3.5"},
    {"Wolf 1055 A", "M3.5"},
    {"Wolf 1055 B", "M5"},
    {"L 674-15", "M4"},
    {"Lalande 27173 A", "K5"},
    {"Lalande 27173 B", "M1"},
    {"Lalande 27173 C", "M3"},
    {"Lalande 27173 D", "T"},
    {"Ross 882", "M4.5"},
    {"CD-40 9712", "M3"},
    {"Eta Cassiopeiae A", "G3"},
    {"Eta Cassiopeiae B", "K7"},
    {"Lalande 46650", "M2"},
    {"36 Ophiuchi A", "K1"},
    {"36 Ophiuchi B", "K1"},
    {"36 Ophiuchi C", "K5"},
    {"CD-36 13940 A", "K3"},
    {"CD-36 13940 B", "M3.5"},
    {"82 Eridani", "G5"},
    {"Delta Pavonis", "G8"},
    {"Wolf 1481", "M3"}
  }; // end nameType

  /** Coordinates X, Y, Z - equatorial frame FK5 - units = light year. */
  public static final double[][] coords = {
    {-1.6085, -1.2309, -3.7021},
    {-1.7295, -1.4473, -3.7781},
    {-1.7295, -1.4473, -3.7781},
    {-0.0568, -5.9198, 0.4867},
    {-7.4463, 2.1176, 0.9528},
    {-6.5210, 1.6470, 4.8806},
    {-1.6220, 8.1357, -2.2669},
    {-1.6220, 8.1357, -2.2669},
    {7.6191, 3.5124, -2.4136},
    {7.6191, 3.5124, -2.4136},
    {1.9346, -8.7628, -3.6561},
    {7.3847, -0.5844, 7.1996},
    {6.2165, 8.3173, -1.5580},
    {8.6336, -2.0767, -6.0234},
    {-10.8733, 0.5841, 0.1520},
    {10.0484, -3.7270, -2.8116},
    {10.0484, -3.7270, -2.8116},
    {10.0484, -3.7270, -2.8116},
    {-4.7705, 10.3125, 1.0407},
    {-4.7705, 10.3125, 1.0407},
    {6.4901, -6.1106, 7.1543},
    {6.4901, -6.1106, 7.1543},
    {1.0925, -5.7821, 10.0431},
    {1.0925, -5.7821, 10.0431},
    {8.3438, 0.6713, 8.0883},
    {8.3438, 0.6713, 8.0883},
    {-6.4217, 8.3841, 5.3308},
    {5.8940, -3.2873, -9.7162},
    {10.3727, 5.0647, -2.8923},
    {5.1434, 7.0792, -8.2990},
    {11.0153, 3.6057, -3.5435},
    {-4.5921, 11.4520, 1.1301},
    {1.8910, 8.8395, -9.0342},
    {7.7817, -6.6874, -7.7693},
    {6.4288, -2.7289, 11.0476},
    {6.4288, -2.7289, 11.0476},
    {-1.7177, 13.3168, -0.2774},
    {-1.7177, 13.3168, -0.2774},
    {-5.2028, -12.6074, -2.7335},
    {-13.7302, -2.0091, 2.2019},
    {-13.7302, -2.0091, 2.2019},
    {11.4055, 0.2688, -8.4883},
    {13.9782, 3.0477, 1.3482},
    {12.2859, 7.1076, 3.2899},
    {-6.7397, 2.3042, -12.7447},
    {-13.7989, 4.4702, -2.7320},
    {-0.5605, -5.4243, 13.7265},
    {-1.4223, -10.3496, -10.4834},
    {-6.7893, 0.4242, -13.4473},
    {15.2260, 0.4452, -1.7265},
    {14.2748, -4.2759, -3.6418},
    {5.2410, -9.6628, 10.7710},
    {5.2410, -9.6628, 10.7710},
    {5.2410, -9.6628, 10.7710},
    {-11.1041, 2.6915, 10.8551},
    {-11.1071, 2.6922, 10.8518},
    {-9.1920, 4.7139, 12.0738},
    {-13.6268, 6.3833, 5.4373},
    {8.4826, -6.2998, -12.1478},
    {7.6659, 10.9075, -9.1856},
    {-1.1867, -11.8368, -11.3616},
    {7.2118, 14.6725, -1.8194},
    {7.2056, 14.6761, -1.8146},
    {7.2056, 14.6761, -1.8146},
    {11.1850, -3.6991, 11.5098},
    {0.3977, -16.5694, 0.7236},
    {0.3977, -16.5694, 0.7236},
    {7.7022, -14.6706, 2.5849},
    {-11.2563, 11.4346, 5.7661},
    {-11.2563, 11.4346, 5.7661},
    {16.6028, 1.1246, -4.7298},
    {16.6028, 1.1246, -4.7298},
    {-0.0076, 17.4906, 0.8248},
    {-3.4467, 0.1852, 17.2480},
    {-15.3275, -7.6170, 4.5488},
    {11.8351, -12.6142, -4.6455},
    {3.5006, 8.5781, 15.4092},
    {3.5006, 8.5781, 15.4092},
    {-3.5625, 14.6140, 9.8682},
    {2.3002, 18.4016, -0.7502},
    {4.5848, -12.4612, -13.0253},
    {-5.1527, -17.8228, -2.4974},
    {-5.1527, -17.8228, -2.4974},
    {-5.1527, -17.8228, -2.4974},
    {2.5644, -6.0121, 17.6379},
    {-0.8174, 17.6605, -6.4814},
    {-0.8174, 17.6605, -6.4814},
    {1.4303, 18.3781, 4.0810},
    {-0.6082, -10.3908, -15.8355},
    {6.2833, -18.0180, 1.7254},
    {6.2914, -18.0157, 1.7199},
    {-9.8441, 15.0577, -6.7083},
    {-12.8878, -12.6096, -6.7712},
    {-12.8933, -12.6040, -6.7712},
    {-12.8933, -12.6040, -6.7712},
    {-12.8946, -12.5943, -6.7870},
    {-8.5192, 17.3324, 1.1981},
    {-8.8161, -11.7207, -12.6224},
    {10.1072, 2.1991, 16.4361},
    {10.1072, 2.1991, 16.4361},
    {19.4313, -0.9164, 0.8153},
    {-3.4099, -17.3003, -8.3728},
    {-3.4099, -17.3003, -8.3728},
    {-3.3395, -17.2606, -8.3667},
    {8.6620, -13.4408, -11.5750},
    {8.6620, -13.4408, -11.5750},
    {9.3089, 11.0841, -13.4663},
    {4.3455, -6.9072, -18.1718},
    {-15.2846, -12.1906, -3.9717}
  }; // end coords
} // end class NearestStarsData